var MAX_SCROLLBACK = 100;

var animatingScroll = false;
var animationInterval = null;
var startScrollTop = 0;
var currentScrollTop = 0;
var targetScrollTop = 0;
var animationComplete = 0;
var contentInset = 10; // Initial top padding is 10px
var autoscrollSuspended = false;

var MESSAGE_FLAGS = { "None":0, "Action":1, "Whisper":1<<1 };
var ADD_MESSAGE_ACTION = { "Append":0, "Prepend":1, "Replace":2 };

function cubicInOut(t, b, c, d) {
    if ((t/=d/2) < 1) return c/2*t*t*t + b;
    return c/2*((t-=2)*t*t + 2) + b;
}

function setClassForOrientation () {  
    var orientationClass = "";
    var orientation = window.orientation;

    if ( orientation === 0 || orientation === 180) {
        orientationClass = "portrait";
    } else if ( orientation === 90 || orientation === -90 ) {
        orientationClass = "landscape";
    }

    document.body.className = orientationClass + " " + setClassForDevice(); 
}

function setClassForOrientationOf (orientation) {  
    var orientationClass = "";

    if ( orientation === 0 || orientation === 180) {
        orientationClass = "portrait";
    } else if ( orientation === 90 || orientation === -90 ) {
        orientationClass = "landscape";
    }

    document.body.className = orientationClass + " " + setClassForDevice(); 
}

window.onorientationchange = setClassForOrientation;

function setClassForDevice() {
    var deviceClass = "";
    if ((navigator.userAgent.match(/iPhone/i)) || (navigator.userAgent.match(/iPod/i))) {
        deviceClass = "iPhone"
    } else if ((navigator.userAgent.match(/iPad/i))) {
        deviceClass = "iPad"
    }
    return deviceClass;
}

function animateScroll(target, duration, callback) {
    var slice = (1000 / 30); // 30 frames per second

    animationComplete = 0;
    startScrollTop = currentScrollTop;
    targetScrollTop = target;

    if (animatingScroll) return;

    animatingScroll = true;

    function step() {
        animationComplete += slice;

        if (animationComplete < duration) {
            currentScrollTop = cubicInOut(animationComplete, startScrollTop, targetScrollTop - startScrollTop, duration);
            document.body.scrollTop = currentScrollTop;
        } else {
            currentScrollTop = targetScrollTop;
            document.body.scrollTop = currentScrollTop;

            clearInterval(animationInterval);
            animationInterval = null;

            animatingScroll = false;
            if (callback) callback();
        }
    }

    step();

    animationInterval = setInterval(step, slice);
}

function stopScrollAnimation() {
    clearInterval(animationInterval);
    animationInterval = null;
    animatingScroll = false;
}

function addComponents(components, action, suppressScroll) {
	
	if (action == ADD_MESSAGE_ACTION.Replace) {
		
		while (document.body.lastChild) {
		
			document.body.removeChild(document.body.lastChild);
		}
	}
	
	if (autoscrollSuspended)
        suppressScroll = true;

	var previousScrollTop = currentScrollTop;
	var previousScrollHeight = document.body.scrollHeight;
    var wasNearBottom = nearBottom();
    var componentsLength = components.length;
    var container = document.createDocumentFragment();

    for (var i = 0; i < componentsLength; ++i) {
        var component = components[i];
        if (component.type === "message") {
            var senderColor = component.senderColor;
            if (senderColor === undefined)
                senderColor = "#000";
			var recipientColor = component.recipientColor;
			if (recipientColor === undefined)
				recipientColor = "#000";

            var badges = component.badges;
            if (badges === undefined)
                badges = "";
			
            appendMessage(container, component.sender, component.recipient, senderColor, recipientColor, component.message, component.flags, component.self, badges);
		}
		else if (component.type === "event") {
            appendEventMessage(container, component.message);
		}
		else if (component.type === "separator") {
			appendSeparator(container, component.message);
		}
    }

	var firstBodyChild = document.body.firstChild;
	if ((action == ADD_MESSAGE_ACTION.Prepend) && firstBodyChild != null) {
	
		document.body.insertBefore(container,firstBodyChild);
	}
	else {
		
		document.body.appendChild(container);
	}

	if (!suppressScroll && (wasNearBottom || action == ADD_MESSAGE_ACTION.Replace)) {
		
		scrollToBottom(false);
	}
	else if (action == ADD_MESSAGE_ACTION.Prepend) {
		
		currentScrollTop = previousScrollTop + (document.body.scrollHeight - previousScrollHeight);
		document.body.scrollTop = currentScrollTop;
	}
}

function badgeMarkup(badges) {
	
	var badge;
	var badgeMarkup = '';
	
	if (badges) {
		
		for (var i = 0; i < badges.length; i++) {
			
			badge = badges[i];
			badgeMarkup += '<img src="' + badge['url'] + '" class="' + badge['type'] + '" />';
		}
	}
	
	return badgeMarkup;
}

function senderMarkup(senderNickname, senderColor, isAction, badges) {
	
	var senderMarkup = '<span class="sender" style="color:' + senderColor + ';">';
	
	if (!isAction) {
		
		senderMarkup += badgeMarkup(badges);
	}
	senderMarkup += (isAction ? "\u2022\u2008 " : '' ) + senderNickname + (isAction ? ' </span>' : ':   </span>');

	return senderMarkup;
}

function whisperMarkup(senderNickname, recipientNickname, senderColor, recipientColor, isAction, badges) {
	
	var senderMarkup = '<span class="sender" style="color:' + senderColor + ';">';
	
	if (!isAction) {
		
		senderMarkup += badgeMarkup(badges);
	}
	senderMarkup += senderNickname + '</span><img src="whisper.svg" class="whisper-to" /><span class="sender" style="color:' + recipientColor + ';">' + recipientNickname + (isAction ? ' </span>' : ':   </span>');
	
	return senderMarkup;
}

function appendMessage(container, senderNickname, recipientNickname, senderColor, recipientColor, messageHTML, flags, self, badges) {
    var badgeElement;
    var className = "message-wrapper";
	var isWhisper = (flags & MESSAGE_FLAGS.Whisper);
	var isAction = (flags & MESSAGE_FLAGS.Action) && !(flags & MESSAGE_FLAGS.Whisper);
	
    if (isAction) className += " action";

    var messageWrapperElement = document.createElement("p");
    messageWrapperElement.className = className;

    var senderWrapperElement = document.createElement("a");
	senderWrapperElement.setAttribute("href","chat://user/" + (self ? recipientNickname : senderNickname));
	if (isWhisper) {
		senderWrapperElement.innerHTML = whisperMarkup(senderNickname, recipientNickname, senderColor, recipientColor, isAction, badges);
	}
	else {
		senderWrapperElement.innerHTML = senderMarkup(senderNickname, senderColor, isAction, badges);
	}
	messageWrapperElement.appendChild(senderWrapperElement);
	
    var messageElement = document.createElement("span");
	if (isWhisper) {
		messageElement.className = "message whisper";
	}
	else {
		messageElement.className = "message " + senderNickname;
	}
    messageElement.innerHTML = messageHTML;
    messageWrapperElement.appendChild(messageElement);

    container.appendChild(messageWrapperElement);
}

function appendEventMessage(container, messageHTML) {
    var className = "message-wrapper event";
    var eventElement = document.createElement("p");
    eventElement.className = className;
    eventElement.innerHTML = messageHTML;

    container.appendChild(eventElement);
}

function appendSeparator(container, title) {
	
	var separatorWrapperElement = document.createElement("h2");
	var separatorTitle = document.createElement("span");
	separatorTitle.innerHTML = title;
	separatorWrapperElement.appendChild(separatorTitle);
	
	container.appendChild(separatorWrapperElement);
}

function replaceMessagesFromUserWithMessage(user, message) {
	var className = "message " + user;

	var elements = document.getElementsByClassName(className);
	for (var i = 0; i < elements.length; i++) {
		elements[i].innerHTML = message;
	}
}

function resetChat() {
	while (document.body.childNodes.length)
		document.body.removeChild(document.body.firstChild);
    scrollToBottom(false, true);
}

function enforceScrollbackLimit() {
    if (document.body.childNodes.length < MAX_SCROLLBACK)
        return;
    while (document.body.childNodes.length > MAX_SCROLLBACK)
        document.body.removeChild(document.body.firstChild);
    scrollToBottom(false, true);
}

function suspendAutoscroll() {
    autoscrollSuspended = true;
}

function resumeAutoscroll() {
    autoscrollSuspended = false;
}

function updateScrollPosition(position) {
    currentScrollTop = position;
}

function nearBottom() {
    return (animatingScroll || currentScrollTop >= (document.body.scrollHeight - contentInset - window.innerHeight - 40));
}

function setContentInset(inset) {
	
	if (inset == contentInset) return;
	
	var insetDelta = inset - contentInset;
	contentInset = inset;
	currentScrollTop = document.body.scrollTop + insetDelta;
	document.body.style.paddingTop = contentInset + "px";
	document.body.scrollTop = currentScrollTop;
}

function scrollToBottomIfNeeded(animated) {
    if (nearBottom())
        scrollToBottom(animated);
}

function scrollToBottom(animated) {
    if (!animated) {
        stopScrollAnimation();

        currentScrollTop = (document.body.scrollHeight - window.innerHeight);
        document.body.scrollTop = currentScrollTop;
        return;
    }

    var newScrollTop = (document.body.scrollHeight - window.innerHeight);
    animateScroll(newScrollTop, 250, animationFinished);
}
